#define lu_decompose
// lu_decompose() - calculates the LU decomposition of a square matrix
// global.lu_n = the size of the matrix
// global.lu_matrix[i, j] = square matrix to decompose
// returns the absolute value of the determinant, which is 0 if the matrix is singular

var i, j, k, pivot, det, sum;

// initialize permutations
for(i = 0; i<global.lu_n; i += 1) {
    global.lu_permutations[i] = i;
}

// choose pivot
pivot = lu_findpivot(0);
if abs(pivot)<0.000000001 {
    return 0;
}
det = pivot;
global.lu_matrix[0, 0] = pivot;

// calculate first column
for(j = 1; j<global.lu_n; j += 1) {
    global.lu_matrix[j, 0] /= pivot;
}

// LU decomposition
for(i = 1; i<global.lu_n; i += 1) {
    
    // choose pivot
    pivot = lu_findpivot(i);
    if abs(pivot)<0.000000001 {
        return 0;
    }
    det *= pivot;
    global.lu_matrix[i, i] = pivot;
    
    // calculate row
    for(j = i+1; j<global.lu_n; j += 1) {
        sum = global.lu_matrix[i, j];
        for(k = 0; k<i; k += 1) {
            sum -= global.lu_matrix[i, k]*global.lu_matrix[k, j];
        }
        global.lu_matrix[i, j] = sum;
    }
    
    // calculate column
    for(j = i+1; j<global.lu_n; j += 1) {
        sum = global.lu_matrix[j, i];
        for(k = 0; k<i; k += 1) {
            sum -= global.lu_matrix[j, k]*global.lu_matrix[k, i];
        }
        global.lu_matrix[j, i] = sum/pivot;
    }
    
}

// the sign of the determinant is incorrect because the rows are swapped
return abs(det);



#define lu_solve
// lu_solve() - solves a system of linear equations based on the LU decomposition of the matrix
// lu_decompose should be called first. Remember to check the return value of lu_decompose!
// global.lu_vector[i] = the vector to solve for
// result: global.lu_result[i]

var i, k, sum;

// create permutated vector
for(i = 0; i<global.lu_n; i += 1) {
    global.lu_result[i] = global.lu_vector[global.lu_permutations[i]];
}

// solve Ly=b for y
for(i = 1; i<global.lu_n; i += 1) {
    sum = global.lu_result[i];
    for(k = 0; k<i; k += 1) {
        sum -= global.lu_matrix[i, k]*global.lu_result[k];
    }
    global.lu_result[i] = sum;
}

// solve Ux=y for x
global.lu_result[global.lu_n-1] /= global.lu_matrix[global.lu_n-1, global.lu_n-1];
for(i = global.lu_n-2; i>=0; i -= 1;) {
    sum = global.lu_result[i];
    for(k = i+1; k<global.lu_n; k += 1) {
        sum -= global.lu_matrix[i, k]*global.lu_result[k];
    }
    global.lu_result[i] = sum/global.lu_matrix[i, i];
}



#define lu_findpivot
// lu_findpivot(i)
// internal - used by lu_decompose

var j, max_j, max_sum, temp;

max_j = argument0;
max_sum = global.lu_matrix[argument0, argument0];
for(k = 0; k<argument0; k += 1) {
    max_sum -= global.lu_matrix[argument0, k]*global.lu_matrix[k, argument0];
}
for(j = argument0+1; j<global.lu_n; j += 1) {
    sum = global.lu_matrix[j, argument0];
    for(k = 0; k<argument0; k += 1) {
        sum -= global.lu_matrix[j, k]*global.lu_matrix[k, argument0];
    }
    if abs(sum)>abs(max_sum) {
        max_j = j;
        max_sum = sum;
    }
}
if max_j!=argument0 {
    for(j = 0; j<global.lu_n; j += 1) {
        temp = global.lu_matrix[argument0, j];
        global.lu_matrix[argument0, j] = global.lu_matrix[max_j, j];
        global.lu_matrix[max_j, j] = temp;
    }
    temp = global.lu_permutations[argument0];
    global.lu_permutations[argument0] = global.lu_permutations[max_j];
    global.lu_permutations[max_j] = temp;
}

return max_sum;



#define lu_decompose_local
// lu_decompose_local() - calculates the LU decomposition of a square matrix
// lu_n = the size of the matrix
// lu_matrix[i, j] = square matrix to decompose
// returns the absolute value of the determinant, which is 0 if the matrix is singular

var i, j, k, pivot, det, sum;

// initialize permutations
for(i = 0; i<lu_n; i += 1) {
    lu_permutations[i] = i;
}

// choose pivot
pivot = lu_findpivot_local(0);
if abs(pivot)<0.000000001 {
    return 0;
}
det = pivot;
lu_matrix[0, 0] = pivot;

// calculate first column
for(j = 1; j<lu_n; j += 1) {
    lu_matrix[j, 0] /= pivot;
}

// LU decomposition
for(i = 1; i<lu_n; i += 1) {
    
    // choose pivot
    pivot = lu_findpivot_local(i);
    if abs(pivot)<0.000000001 {
        return 0;
    }
    det *= pivot;
    lu_matrix[i, i] = pivot;
    
    // calculate row
    for(j = i+1; j<lu_n; j += 1) {
        sum = lu_matrix[i, j];
        for(k = 0; k<i; k += 1) {
            sum -= lu_matrix[i, k]*lu_matrix[k, j];
        }
        lu_matrix[i, j] = sum;
    }
    
    // calculate column
    for(j = i+1; j<lu_n; j += 1) {
        sum = lu_matrix[j, i];
        for(k = 0; k<i; k += 1) {
            sum -= lu_matrix[j, k]*lu_matrix[k, i];
        }
        lu_matrix[j, i] = sum/pivot;
    }
    
}

// the sign of the determinant is incorrect because the rows are swapped
return abs(det);



#define lu_solve_local
// lu_solve_local() - solves a system of linear equations based on the LU decomposition of the matrix
// lu_decompose should be called first. Remember to check the return value of lu_decompose!
// lu_vector[i] = the vector to solve for
// result: lu_result[i]

var i, k, sum;

// create permutated vector
for(i = 0; i<lu_n; i += 1) {
    lu_result[i] = lu_vector[lu_permutations[i]];
}

// solve Ly=b for y
for(i = 1; i<lu_n; i += 1) {
    sum = lu_result[i];
    for(k = 0; k<i; k += 1) {
        sum -= lu_matrix[i, k]*lu_result[k];
    }
    lu_result[i] = sum;
}

// solve Ux=y for x
lu_result[lu_n-1] /= lu_matrix[lu_n-1, lu_n-1];
for(i = lu_n-2; i>=0; i -= 1;) {
    sum = lu_result[i];
    for(k = i+1; k<lu_n; k += 1) {
        sum -= lu_matrix[i, k]*lu_result[k];
    }
    lu_result[i] = sum/lu_matrix[i, i];
}



#define lu_findpivot_local
// lu_findpivot_local(i)
// internal - used by lu_decompose_local

var j, max_j, max_sum, temp;

max_j = argument0;
max_sum = lu_matrix[argument0, argument0];
for(k = 0; k<argument0; k += 1) {
    max_sum -= lu_matrix[argument0, k]*lu_matrix[k, argument0];
}
for(j = argument0+1; j<lu_n; j += 1) {
    sum = lu_matrix[j, argument0];
    for(k = 0; k<argument0; k += 1) {
        sum -= lu_matrix[j, k]*lu_matrix[k, argument0];
    }
    if abs(sum)>abs(max_sum) {
        max_j = j;
        max_sum = sum;
    }
}
if max_j!=argument0 {
    for(j = 0; j<lu_n; j += 1) {
        temp = lu_matrix[argument0, j];
        lu_matrix[argument0, j] = lu_matrix[max_j, j];
        lu_matrix[max_j, j] = temp;
    }
    temp = lu_permutations[argument0];
    lu_permutations[argument0] = lu_permutations[max_j];
    lu_permutations[max_j] = temp;
}

return max_sum;



